import argparse
import socket
import sys
import signal

try:
    import queue
except ImportError:
    import Queue as queue

from elasticsearch_raven import configuration
from elasticsearch_raven import transport
from elasticsearch_raven import queue_sender
from elasticsearch_raven import queues
from elasticsearch_raven import udp_handler


def run_server():
    args = _parse_args()
    try:
        sock = get_socket(args.ip, args.port)
    except socket.gaierror:
        sys.stdout.write('Wrong hostname.\n')
        sys.exit(1)
    else:
        log_transport = transport.get_configured_log_transport()
        if args.amqp_queue:
            pending_logs = queues.KombuQueue(configuration['amqp_url'],
                                             configuration['amqp_queue'])
        else:
            pending_logs = queues.ThreadingQueue(
                configuration['queue_maxsize'])
        Server(sock, pending_logs, log_transport,
               args.debug).run()


def _parse_args():
    parser = argparse.ArgumentParser(description='Udp proxy server for raven')
    parser.add_argument('ip')
    parser.add_argument('port', type=int)
    parser.add_argument('--debug', dest='debug', action='store_const',
                        const=True, default=False,
                        help='Print debug logs to stdout')
    parser.add_argument(
        '--amqp-queue', action='store_const', const=True, default=False,
        help='Use amqp queue to store logs to send to elasticsearch.')
    return parser.parse_args()


def get_socket(ip, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((ip, int(port)))
    return sock


class Server(object):
    def __init__(self, sock, pending_logs, log_transport, debug=False):
        self.sock = sock
        self.pending_logs = pending_logs
        self.log_transport = log_transport
        self.debug = debug
        self.exception_queue = queue.Queue()

    def run(self):
        handler = udp_handler.Handler(
            self.sock, self.pending_logs, self.thread_exception_handler,
            debug=self.debug)
        sender = queue_sender.Sender(self.log_transport, self.pending_logs,
                                     self.thread_exception_handler)
        handler.as_thread().start()
        sender.as_thread().start()

        def terminate(signum, frame):
            self.exception_queue.put(KeyboardInterrupt())

        signal.signal(signal.SIGTERM, terminate)
        signal.signal(signal.SIGQUIT, terminate)
        try:
            raise self.exception_queue.get()
        except KeyboardInterrupt:
            handler.should_finish = True
            try:
                while self.pending_logs.has_nonpersistent_task():
                    try:
                        raise self.exception_queue.get(timeout=1)
                    except queue.Empty:
                        pass
            except KeyboardInterrupt:
                pass

    def thread_exception_handler(self, exception):
        self.exception_queue.put(exception)



import signal
import threading

import elasticsearch

from elasticsearch_raven import configuration
from elasticsearch_raven import utils


class Sender(object):
    def __init__(self, log_transport, pending_logs, exception_handler):
        self.log_transport = log_transport
        self.pending_logs = pending_logs
        self.exception_handler = exception_handler

    def as_thread(self):
        sender = threading.Thread(target=self.send)
        sender.daemon = True
        return sender

    def send(self):
        try:
            while True:
                message = self.pending_logs.get()
                self._send_message(message)
        except Exception as e:
            self.exception_handler(e)

    def _send_message(self, message):
        for retry in utils.retry_loop(1.0, max_delay=60.0, back_off=1.5):
            with utils.ignore_signals([signal.SIGTERM, signal.SIGQUIT]):
                try:
                    self.log_transport.send_message(message)
                except elasticsearch.exceptions.ConnectionError as e:
                    retry(e)
                except elasticsearch.exceptions.TransportError as e:
                    self._raport_error(message, e)
                    self.pending_logs.task_done()
                else:
                    self.pending_logs.task_done()

    def _raport_error(self, message, error):
        connection = elasticsearch.Elasticsearch(
            hosts=[configuration['host']],
            use_ssl=configuration['use_ssl'],
            http_auth=configuration['http_auth'])
        body = {'message': str(message), 'error': str(error)}
        connection.index(index='elasticsearch-raven-error', body=body,
                         doc_type='elasticsearch-raven-log')



import datetime
import sys
import signal
import threading

from elasticsearch_raven import transport
from elasticsearch_raven import utils


class Handler(object):
    def __init__(self, sock, pending_logs, exception_handler, debug=False):
        self.sock = sock
        self.pending_logs = pending_logs
        self.exception_handler = exception_handler
        self.debug = debug
        self.should_finish = False

    def as_thread(self):
        handler = threading.Thread(target=self.handle)
        handler.daemon = True
        return handler

    def handle(self):
        try:
            try:
                while True:
                    data, address = self.sock.recvfrom(65535)
                    with utils.ignore_signals([signal.SIGTERM, signal.SIGQUIT]):
                        message = transport.SentryMessage.create_from_udp(data)
                        self.pending_logs.put(message)
                    if self.debug:
                        sys.stdout.write('{host}:{port} [{date}]\n'.format(
                            host=address[0], port=address[1],
                            date=datetime.datetime.now()))
            finally:
                self.sock.close()
        except Exception as e:
            self.exception_handler(e)



import contextlib
import time
import signal


def retry_loop(delay, *, max_delay=None, back_off=1.0):
    exceptions = set()
    if max_delay is None:
        max_delay = delay

    def retry(exception):
        exceptions.add(exception)
    yield retry

    while delay < max_delay:
        if not exceptions:
            return
        time.sleep(delay)
        delay *= back_off
        exceptions.clear()
        yield retry

    while True:
        if not exceptions:
            return
        time.sleep(max_delay)
        exceptions.clear()
        yield retry


@contextlib.contextmanager
def ignore_signals(signals):
    signal.pthread_sigmask(signal.SIG_BLOCK, signals)
    yield
    signal.pthread_sigmask(signal.SIG_UNBLOCK, signals)



import socket
import signal

try:
    from urllib import parse
except ImportError:
    import urlparse as parse

import argparse

from elasticsearch_raven import configuration
from elasticsearch_raven import transport
from elasticsearch_raven import queues
from elasticsearch_raven import queue_sender
from elasticsearch_raven import udp_handler


def run_handler():
    args = _parse_handler_args()
    parsed = parse.urlparse(args.listen_address)
    if parsed.scheme == 'fd':
        sock = socket.fromfd(int(parsed.netloc), socket.AF_INET,
                             socket.SOCK_DGRAM)
    elif parsed.scheme == 'udp':
        ip, port = parsed.netloc.split(':')
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind((ip, int(port)))
    else:
        raise ValueError('only fd and udp protocols are supported')
    sock.setblocking(1)
    sock.settimeout(None)
    queue = queues.KombuQueue(configuration['amqp_url'],
                              configuration['amqp_queue'])
    handler = udp_handler.Handler(sock, queue, _exception_handler,
                                  debug=args.debug)

    def terminate(signum, frame):
        exit(0)
    signal.signal(signal.SIGTERM, terminate)
    signal.signal(signal.SIGQUIT, terminate)
    handler.handle()


def _parse_handler_args():
    parser = argparse.ArgumentParser(description='Udp proxy server for raven')
    parser.add_argument('listen_address', help='''Accepted values are fd://FD or udp://IP:PORT''')
    parser.add_argument('--debug', dest='debug', action='store_const',
                        const=True, default=False,
                        help='Print debug logs to stdout')
    return parser.parse_args()


def run_sender():
    log_transport = transport.get_configured_log_transport()
    queue = queues.KombuQueue(configuration['amqp_url'],
                              configuration['amqp_queue'])
    sender = queue_sender.Sender(log_transport, queue, _exception_handler)

    def terminate(signum, frame):
        exit(0)
    signal.signal(signal.SIGTERM, terminate)
    signal.signal(signal.SIGQUIT, terminate)
    sender.send()


def _exception_handler(exception):
    raise exception


